/*{{{  Change history*/
/****************************************************************************
 * Written by Allen Knutson [allenk@through.ugcs.caltech.edu]        6/88   *
 *  Things to be added:                                                     *
 *      Several patterns listed on screen, with switching between them.     *
 *      Integrate the pattern maker LISTALLP that's in /src/games.          *
 *                                                                          *
 * Notes:                                                                   *
 *         I'm assuming that hold time = time of a 1-throw; theta = 1/2.    *
 *         Thus n-throw + hold time =   n      * (1-throw + hold time),     *
 *                                             as n-throws are defined,     *
 *              n-throw             = (2n-1)   *  1-throw.                  *
 *         Thus n-throw height      = (2n-1)~2 *  1-throw height,           *
 *                                   inducing the calculation above.        *
 ****************************************************************************/
 
/****************************************************************************
 *                                                                          *
 * Various mods by J. Gaudreau:                                             *
 *  7-May-1991 : Reformatted everything to make it "consistent."            *
 * 30-Jun-1991 : Added eXtensions.                                          *
 *                                                                          *
 ****************************************************************************/
 
/**************************************************************************
                                                                         
 Viciously hacked by Alastair Reid <areid@uk.ac.glasgow.dcs>             
 16-Sep-1991 : Keyboard operation partially restored                     
               Default value of detail adjusted to suit speed of SPARC   
               Slight changes made to use "virtual window"               
 17-Sep-1991 : Variable manipulation extensively modified because I      
               couldn't understand it.                                   
               Maximum value of detail adjusted to suit SPARC            
 18-Sep-1991 : Adapted towards my porgramming style and comments added   
               so that I can understand it.                              
               Major stylistic changes:                                  
                 I like 0..i-1 instead of 1..i                           
                 I dislike ++, --, -=, etc.                              
                 I dislike WITH=&some_var ... WITH->x                    
 21-Sep-1991 : Replaced part of animation routine with new parabola
               routine that allows for different hand heights. Gravity
               introduced to replace origina; scaling method. (Gravity
               probably needs more work to avoid arithmetic overflow.)
 23-Sep-1991 : Introduced hands explicitly - allowing them to be at
               different heights.
 10-Oct-1991 : Fixed a bug I'd added: startup sequences of odd length
               resulted in balls instantaneously switching hands shortly 
               after the end of the startup.
                                                                         
***************************************************************************/

/***************************************************************************
Lovingly hacked by John Robinson <jr@ksr.com>

 spring 1991 : added support for running with Unix curses(3) for dumb
               terminals.
  7-Nov-1991 : merged IBMPC, X and curses versions into a common source.
               formatted everything; diff'ing is hopeless now.
  8-Nov-1991 : add pause command: p or P.
 19-Nov-1991 : added private sleep routine jsleep() for Unix variants
               fixed initial display (showed balls before they were placed)
  3-Dec-1991 : added variable number of hands (1-10), set by -h switch
               made scoop per-hand; reduce initial value for large patterns
               added interactive step-to-next command: n/N
               for nhands != 2, fiddle last stepped-to hand only
               added -n switch for curses version to juggle "numbers";
                  default is letter O
               tweaked parabola() to avoid overflow for large numbers of balls
	          and/or large detail; thanks to jvert@microsoft.com
               added checking of patterns, plus automatic generation of
	          startup patterns if needed and none is provided
               made manifest constants for some of the hard-coded values
***************************************************************************/
/*}}}  */

/*{{{  Includes*/
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>

#include "3d.h"
#include "animate.h"
#include "juggle.h"
#include "siteswap.h"
#include "x_uif.h"

#define extern
#include "globals.h"

#include "polyinfo.h"
/*}}}  */
/*{{{  Globals*/
int      m_down, m_across; /* size of "virtual window" */

ball     balls[MAX_PATTERN_LENGTH]; /* ball[0..] = state of balls */
long     now, ticks, throwsnow;

boolean  emptyhand;
long     sum;
boolean  diagnos = 0;
int      ch;

int      gravity = -1;
int      detail;
int      height;

siteswap *cur_ss = NULL;
siteswap *new_ss;

boolean  waiting;

#define  OBJ_SIZE  16

/*{{{  The robot body*/
#define BODY_WIDTH  640
#define BODY_HEIGHT 700

#define HEAD_SIZE   200

#define LEG_LENGTH  600

Polyinfo   body_poly[] = {
#include "data/objcube.h"
#include "data/objbody.h"
#include "data/objoctcyl.h"
};
#define  HEAD_P   0
#define  BODY_P   1
#define  LEG_P    2

#define  NUM_BODY_PARTS 4
anim_pix body_part[NUM_BODY_PARTS];

#define  HEAD_I       0
#define  BODY_I       1
#define  LEFT_LEG_I   2
#define  RIGHT_LEG_I  3
int      body_ind[] = {HEAD_P,BODY_P,LEG_P,LEG_P};
/*}}}  */

/*{{{  Robot arms*/
#define UPPER_ARM   300
#define LOWER_ARM   300


#define  LOW_ARM_POLY 0
#define  UPP_ARM_POLY 1

Polyinfo arm_poly[] = {
#include "data/objlowarm.h"
#include "data/objuparm.h"
};

/*}}}  */
/*}}}  */
/*{{{  Defines*/
#define STD_JTIME 6000   
/* Time (in micro-seconds) between catches 
 * with a 'height' value of 1 
 */

#define THROW_IS_HOLD(t)  ((t == nhands) && (nhands <= 2))

#define sign(b) ( b>0 ? 1: -1 )

#define round(x) (x>0 ? (int)(x+0.5) : (int)(x-0.5))

#define strtol(a,b,c) atoi(a)
/*}}}  */

void do_cube(point *nll, int side)
/*{{{  */
{
   int   x = nll->x,
         y = nll->y,
         z = nll->z;
   point p1,p2;
   
   p2.x = x; p2.y = y+side; p2.z = z;
   draw_line(nll,&p2);
   p1.x = x; p1.y = y+side; p1.z = z-side;
   draw_line(&p1,&p2);
   p1.x = x+side; p1.y = y+side; p1.z = z;
   draw_line(&p1,&p2);
   p2.x = x+side; p2.y = y; p2.z = z;
   draw_line(&p1,&p2);
   
   p1.x = x+side; p1.y = y+side; p1.z = z-side;
   p2.x = x+side; p2.y = y+side; p2.z = z;
   draw_line(&p1,&p2);
   p2.x = x+side; p2.y = y; p2.z = z-side;
   draw_line(&p1,&p2);
   p2.x = x; p2.y = y+side; p2.z = z-side;
   draw_line(&p1,&p2);
   p1.x = x; p1.y = y; p1.z = z-side;
   draw_line(&p1,&p2);
}
/*}}}  */

void redo_screen(void)
/*{{{  */
{
   int   i,x;
   int   side  = min(m_across,m_down);
   int   side2 = side/2;
   point p1,p2;
   
   clrscr();
   
   if (show_grid)
   /*{{{  Draw floor grid*/
   {
      p1.y = p2.y = -side2;
      for (i=-side2; i<=side+side2; i+=side/40)
      {
         p1.z = side;
         p2.z = -side;
         p1.x = p2.x = i;
         draw_line(&p1, &p2);
         
         p1.x = -side2;
         p2.x = side+side2;
         p1.z = p2.z = i-side2;
         draw_line(&p1, &p2);
      }
   }
   /*}}}  */
   
   if (show_cube)
   /*{{{  Draw cube*/
   {   
      p1.x = -side2;
      p1.y = -side2;
      p1.z = side;
      do_cube(&p1, side*2);
   }
   /*}}}  */
   
   if (show_arms)
   /*{{{  */
   {
      /*{{{  Line version*/
      /*
      for(i=0; i<nhands; i++)
      {
        draw_line(&hands[i].shoulder, &hands[i].elbow);
        draw_line(&hands[i].elbow,    &hands[i].wrist);
      }
      */
      /*}}}  */
      
      for (i=0; i<nhands; i++)
      {
         update_ap(&hands[i].up_ap,  &arm_poly[UPP_ARM_POLY], -1);
         draw_pix(hands[i].up_ap.pix);
         
         update_ap(&hands[i].low_ap, &arm_poly[LOW_ARM_POLY], -1);
         draw_pix(hands[i].low_ap.pix);
      }
   }
   /*}}}  */
    
   if (show_robot)
   /*{{{  */
      for(i=0; i<NUM_BODY_PARTS; i++)
      {
         update_ap(&body_part[i], &body_poly[body_ind[i]], -BODY_P); 
         draw_pix(body_part[i].pix);
      }
   /*}}}  */
    
   for(i=0; i<MAX_NUM_HANDS+1; i++)
      if (display_point[i] != NULL)
      /*{{{  */
      {
         p1 = p2 = *display_point[i];
         p1.x -= 20;
         p2.x += 20;
         draw_line(&p1, &p2);   
         p1 = p2 = *display_point[i];
         p1.y -= 20;
         p2.y += 20;
         draw_line(&p1, &p2);   
         p1 = p2 = *display_point[i];
         p1.z -= 20;
         p2.z += 20;
         draw_line(&p1, &p2);   
      }
      /*}}}  */
   
   for (i = 0; i < cur_ss->balls; i++) 
   {
      update_ap(&balls[i].ap, NULL, i);
      draw_pix(balls[i].ap.pix);
   }
   
   cleardevice();
}
/*}}}  */

long parabola(long a, long dy, long t, long T)
/*{{{  */
/* returns height of ball at time t/T thrown on path satisfying the
   following constraints.

   (0) y(t) = v * (t/T) + 0.5 * a * sqr(t/T)
       (ball thrown with initial velocity v in gravitational field a)
   (1) y(T) = dy
       (ball lands at height dy at time 1)

   Solving for v, we get

     y(T) = dy
   = {constraint 0}
     v * (T/T) + 0.5 * a * sqr(T/T) = dy
   = {arithmetic}
     v + 0.5 * a = dy
   = {algebra}
     v = dy - 0.5 * a

   This solution can be used to produce a calculation that strikes a
   balance between roundoff errors and overflows.

     y(t)
   = {constraint 0}
     v * (t/T) + 0.5 * a * sqr(t/T)
   = {v = dy - 0.5 * a}
     (dy - 0.5 * a) * (t/T) + 0.5 * a * sqr(t/T)
   = {arithmetic}
     ((2 * dy - a) * t * T + (a * sqr(t))) / (2 * sqr(T))

*/
/* Note that for dy = 0, y(t) is a*(t-T)*t / (2 * sqr(T)).
   The original code (which ignored dy) uses the following calculation

     (T-t) * t * 4 / sqr(detail) * scale/KLUDGE		[KLUDGE == 100]

   This suggests that an accelaration proportional to -sqr(throw) is
   appropriate.

from jvert@microsoft.com:
    return  t * ((a - 2 * dy) - (a * t) / T) / (2 * T);

that is close, but above rewrites as [opposite sign from jvert]:
   = {arithmetic}
     ((t * ((2 * dy - a)  + (a * t) / T)) / 2) / T

*/

/*{{{  COMMENT - old version*/
/* this calculation replaced by use of parabola */
/* long y_flight(long throw, long lat, long j)
/* /* j / detail == how far between sites we are */
/* /* throw = flight time 
/*    lat = amount of flight done */
/* {
/*   long howfar, throwfactor;
/* 
/*   howfar = (lat - 1) * detail + j;
/*   throwfactor = throw - 1;
/*   return( (scale*throwfactor - scale*howfar/detail) * howfar*4/detail/KLUDGE );
/* }
*/
/*}}}  */

{ return( ((t * ((2 * dy - a)  + (a * t) / T)) / 2) / T ); }
/*}}}  */

long parabola2(long dy, long t, long T)
/*{{{  */
{
   return (t * (dy/T - t + T)) >> 3;
}
/*}}}  */

void calc_ball_position(int ball_i, long j)
/*{{{  */
{
    ball   *b      = &balls[ball_i];
    hand   *hf     = &hands[b->handfrom],
           *ht     = &hands[b->handto];
    int    lateral = b->lateral,
           throw   = b->throw,
           throw2  = 2*throw,
           dw      = min(ht->dwell,throw2-1),
           xscoop  = ht->throw.x - ht->catch.x,
           yscoop  = ht->throw.y - ht->catch.y,
           zscoop  = ht->throw.z - ht->catch.z;
    int    tDir    = sign(ht->catch.x - ht->throw.x),
           fDir    = sign(hf->catch.x - hf->throw.x);
    long   t,T;
    long   dx,dy,dz;
    long   xdisp, ydisp, zdisp;
    int    num_spins;
    double spin, splay;
    point  pos;
    
    /*{{{  Comments*/
    /* b = ball
       j / detail == how far between sites we are
       2 * b->throw == flight time  (ie b->throw == height of throw)
       b->lateral == amount of flight done
       return new positions in x and y
    */
    /*}}}  */
    
    
    if (throw <=0)
    /*{{{  Ball is in hand */
    {		
    	dx = - b->left_right	/* Negate because lateral is <=0 */
	    * (3 * OBJ_SIZE * ((lateral - 1) * detail + j))/detail;
        dy = 0;
    	dz = 0;
    
        spin  = 0.0;
        splay = 0.0;
    } 
    /*}}}  */
    else 
    
    if ( lateral > throw2 - dw ) 
    /*{{{  Scooping */
    {
        t = (lateral - throw2 + dw - 1)*detail + j; /* amount of scoop so far */
        T = dw*detail;                              /* total scoop time */
        
        dx = ht->catch.x - hf->throw.x + xscoop*t/T;
        dy = ht->catch.y - hf->throw.y + parabola( height*throw2, yscoop, t, T);
        dz = ht->catch.z - hf->throw.z + zscoop*t/T;
    
        spin  = PI_BY_2 * (double)dy / (50*throw2);
        splay = (fDir + (double)t/T * (tDir - fDir) ) * PI_BY_4;
    } 
    /*}}}  */
    else 

    if ( THROW_IS_HOLD(throw) ) 
    /*{{{  Special case: hold in hand */
    {
        /* Well, wave it about a bit actually */
        
        t = (lateral - 1)*detail + j;
        T = (4 - dw)*detail;
        
        dx = xscoop * t / T;
    	dy = 0;
    	dz = 0;
    
        spin  = 0.0;
        splay = fDir * PI_BY_4;
    } 
    /*}}}  */
    else 
    
    /*{{{  In flight */
    {			
        xdisp = ht->catch.x - hf->throw.x;
    	ydisp = ht->catch.y - hf->throw.y;
        zdisp = ht->catch.z - hf->throw.z;
        
        t = (lateral - 1)*detail + j; /* amount of flight so far */
    	T = (throw2 - dw)*detail;     /* total flight time */
      
    	dx = (xdisp * t)/T;
    	dz = (zdisp * t)/T;
        dy = parabola2( ydisp, t*height/detail, T*height/detail );
        
        spin  = (((double)t) * TWO_PI)/T;
        splay = fDir * PI_BY_4;
    }
    /*}}}  */
    
    pos.x = hf->throw.x + dx;
    pos.y = hf->throw.y + dy;
    pos.z = hf->throw.z + dz;
    
    num_spins = (b->throw<2) ? 0 : (int)sqrt((double)(b->throw));
    spin *= num_spins;
    
    set_ball_tx(b->ap.tx, spin, splay, &pos);
    update_ap(&b->ap, NULL, ball_i);
}
/*}}}  */

void calc_ball_state(int ball_i, int throw_i)
/*{{{  */
{
   ball *b       = &balls[ball_i];
   int  thrownow = cur_ss->pattern[throw_i];
   int  throw    = max(0, b->throw);
   int  throw2   = throw*2;
   int  dw       = hands[b->handto].dwell;
   int  i;
   
   /*{{{  Comments*/
   /* 
      Updates count "throwsnow" of total number of throws so far
      
      Updates flag "emptyhand" to watch for hands being overfilled
      
      It might be useful to extend meaning of "thrownow" in something
      like the following way:
      -1 : pickup ball
      -2 : drop ball
   */
   /*}}}  */
   
   if (b->throw == 0)
      printf("****** Throw = 0 ***********\n");
   
   if (b->lateral < throw2) 
   {                       /* Approaching hand */
      b->lateral++;	   /* nudge it on */
   
      /* Ball is being scooped by a hand */
      if (b->lateral > throw2 - (throw==0 ? dw : min(dw, throw2-1)) )
      {
         hands[b->handto].ball = b;
      }

      /*{{{  Pause stuff*/
      if (b->lateral == b->throw*2) 
      {
         if (next_pause)  /* this is an interesting place to pause */
         {
            next_pause = 0;
            pausing    = 1;
         }
      }
      /*}}}  */
   } 
   else 
   {                      /* Ball ready to leave hand */
      /*{{{  Pause stuff*/
      if (next_pause) 
      {	/* this is an interesting place to pause */
         next_pause = 0;
         pausing = 1;
      }
      /*}}}  */
              
      if ( !THROW_IS_HOLD(thrownow) )
      {
         hands[b->handto].ball = NULL;
         hands[b->handto].lateral = 0;
         i = ((throw_i-MAX_PATTERN_LENGTH)+nhands-1)%cur_ss->pattern_len;
         hands[b->handto].prior_throw = cur_ss->pattern[MAX_PATTERN_LENGTH + i];
      }
      
      b->lateral  = 1;
      b->throw    = thrownow;
      b->handfrom = b->handto; /* ball is where it last landed */
      b->handto   = (b->handfrom + b->throw) % nhands;
       
      throwsnow++;  
   }

   return;
}
/*}}}  */

void calc_hand_state(int hand_i,int throw_i)
/*{{{  */
{
   hand *h    = &hands[hand_i];
   int  found = 0,
        t     = throw_i - MAX_PATTERN_LENGTH;
   
   if (h->ball != NULL) return;
   
   if (h->lateral == 0)
   {
      h->recoil = 0;
      while( !found && h->recoil < 100)
      {
         t = (t+nhands)%cur_ss->pattern_len;
         h->recoil += 2*nhands;
         found = cur_ss->pattern[MAX_PATTERN_LENGTH + t] != 0;
      }
   }
   
   h->lateral++;

}
/*}}}  */

void calc_elbow_position(hand *h)
/*{{{  */
{
   point  d, e; 
   double H, H2, Y2, K, a, b, c, rt, hl, hu, yl, yu, sina, cosa;
   tx     *utx = h->up_ap.tx,
          *ltx = h->low_ap.tx;
   
   sub_point(&h->shoulder, &h->wrist, &d);
   
   H2 = sqr(d.x) + sqr(d.z);
   H  = sqrt(H2);
   Y2 = sqr(d.y);
   K  = sqr(UPPER_ARM) - (sqr(LOWER_ARM) + H2 + Y2);
   a  = (Y2 + H2);
   b  = K * H;
   c  = sqr(K)/4 - Y2*sqr(LOWER_ARM);
   rt = sqr(b) - 4*a*c;
   
   if (rt < 0)
   {
      /* Not sure if this is a good idea. What we want is
       * to stretch as far in the right direction as possible
       */
      rt = 0;
   }
   
   hl = (-b + sqrt(rt))/(2 * a);
   hu = H - hl;
   
   if (hu < 0)
   {
      /* printf("hu = %lf!\n",hu);*/
      hu = 0.0;
   }
   yu = sqr(UPPER_ARM) - sqr(hu);
   if (yu < 0)
   {
      /*printf("yu2 = %lf!\n",yu);*/
      yu = 0.0;
   }
   yu = sqrt(yu);
   yl = yu - d.y;
      
   sina =  (double)d.x/H;
   cosa =  (double)d.z/H;   
   
   h->elbow.x = h->wrist.x + hl * sina;
   h->elbow.y = h->wrist.y - yl;
   h->elbow.z = h->wrist.z + hl * cosa;
      
   /* Rotate upper arm down, then across
    * and shift over to start at shoulder
    */
   identity_tx(utx);
   rotate_tx_sc(utx, 'x', -yu/UPPER_ARM, hu/UPPER_ARM);
   rotate_tx_sc(utx, 'y', sina, cosa);
   offset_tx(utx, &h->shoulder);
   
   /* And bring lower arm up, then across,
    * shift to elbow position
    */
   identity_tx(ltx);
   rotate_tx_sc(ltx, 'x', yl/LOWER_ARM, hl/LOWER_ARM);
   rotate_tx_sc(ltx, 'y', sina, cosa);
   offset_tx(ltx, &h->elbow);
   
   update_ap(&h->up_ap,  &arm_poly[UPP_ARM_POLY], -1);
   update_ap(&h->low_ap, &arm_poly[LOW_ARM_POLY], -1);
}
/*}}}  */

void calc_wrist_position(hand *h, long j)
/*{{{  */
{
   if (h->ball != NULL)
   {
      tx *tx = h->ball->ap.tx;
      
      h->wrist.x = tx->dx;
      h->wrist.y = tx->dy;
      h->wrist.z = tx->dz;
   }
   else
   {
      int xscoop = h->throw.x - h->catch.x,
          yscoop = h->throw.y - h->catch.y,
          zscoop = h->throw.z - h->catch.z;
      int t      = (h->lateral - 1)*detail + j; 
      int dw     = min(h->dwell, h->prior_throw*2 - 1);
      int T      = (h->recoil - dw)*detail;

      if (T == 0)
         t = T = 1;
         
      h->wrist.x = h->throw.x - xscoop * t / T;
      h->wrist.y = h->throw.y - parabola( 3*abs(xscoop), yscoop, t, T);
      h->wrist.z = h->throw.z - zscoop * t / T;
   }
   
}
/*}}}  */

void do_diagnostics()
/*{{{  */
{
    int  i;
    char msg[256];
    
    msg[0] = 0;
    
    if (diagnos) {
	printf("Ticks %2d Now=%2d ",ticks,now);
        for (i = 0; i<cur_ss->balls; i++) 
	   printf(" %2d/%2d", balls[i].throw, balls[i].lateral);
        printf(" [");
        for(i=0; i<nhands; i++)
           printf(" %c%2d/%2d",
                  hands[i].ball==NULL? 'P' : 'B',
                  hands[i].ball==NULL?hands[i].recoil : (hands[i].ball-&balls[0]),
                  hands[i].lateral);
        printf("]\n");
    }
    
    if (throwsnow != (ticks + 1 & 1) && 
       !(throwsnow == 0 && 
       cur_ss->pattern[now+MAX_PATTERN_LENGTH] == 0)) 
    {
	sprintf(msg+strlen(msg),"\n     Bad pattern: caught %ld balls.", throwsnow); 
        diagnos = 1;
    }
    if (emptyhand) {
	sprintf(msg+strlen(msg),"\n     Bad pattern: was holding a ball during a 0.");
        diagnos = 1;
    }
    
    if (diagnos && msg[0])
       printf(msg);

}
/*}}}  */

void parse_args(int argc, char *argv[], siteswap *ss)
/*{{{  */
{
    /*
    if (argc == 1) {
	printf("Usage: %s [-n] [-h <hands>] <pattern> [-s <pattern>]\n",argv[0]);
	exit(-1);
    }
    */
    argc--; argv++; /* Skip program name */
    
    show_grid = show_cube = FALSE;
    show_arms = show_robot = TRUE;

    done       = 0;
    pausing    = 0;
    next_pause = 0;
    onacid     = 0;
        
    ss->pattern_str[0] = '\0';
    ss->pattern_len    = 0;
    ss->startup_str[0] = '\0';
    ss->startup_len    = 0;
    ss->hands          = 2;
    ss->valid          = FALSE;

    /*{{{  COMMENT do_numbers stuff*/
    /* CURSES: juggle numbers, not O's */
    /*
    if (argc > 0 && !strcmp(argv[0], "-n")) {
    	do_numbers = 1;		
    	argc--; argv++;
    } else
    	do_numbers = 0;
    */
    /*}}}  */
    
    if (argc >= 2 && !strcmp(argv[0], "-h")) 
    {
	ss->hands = atoi(argv[1]);
	argc--; argc--; argv++; argv++;
	
        if (ss->hands <= 0 || ss->hands > MAX_NUM_HANDS) 
        {
	    printf("Improper number of hands: %d\n", ss->hands);
	    exit(-1);
	}
    }

    while( (argc>0) && (argv[0][0] != '-') )
    {
       strcat(ss->pattern_str, argv[0]);
       strcat(ss->pattern_str, " ");
       ss->pattern_len++;
       argc--;
       argv++;
    }
    
    if ( (argc>0) && (strcmp(argv[0], "-s") == 0) )
    {
       argc--; argv++;
       
       while( argc>0 )
       {
          strcat(ss->startup_str, argv[0]);
          strcat(ss->startup_str, " ");
          ss->startup_len++;
          argc--;
          argv++;
       }
    }
    
    if (ss->pattern_len == 0)
    {
       fprintf(stderr,"No site-swap pattern given - defaulting to 3\n");
       strcpy(ss->pattern_str,"3");
       ss->pattern_len = 1;
    }
   
    /*{{{  debugx*/
    if ( (argc>0) && (strcmp(argv[0], "-debugx") == 0) )
    {
       debug_x = TRUE;
       argc--;
       argv++;
    }
    else
       debug_x = FALSE;
    /*}}}  */
    /*{{{  diagnos*/
    if ( (argc>0) && (strcmp(argv[0], "-diagnos") == 0) )
    {
       diagnos = TRUE;
       argc--;
       argv++;
    }
    else
       diagnos = FALSE;
    /*}}}  */
    
    extract_pattern(ss);
}
/*}}}  */

void
process_events()
/*{{{  */
{
   JballEvent();
   
   if (pausing) 
     jsleep(30000); /* 30 milliseconds */
}
/*}}}  */

void load_pattern(siteswap *ss)
/*{{{  */
{
   siteswap *prev_ss = cur_ss;
   int      i;
   
   if (ss->hands != nhands)
   /*{{{  set initial hand positions*/
   {
      int  scoop = min(m_across/10 + (10 * ss->balls), m_across/3);
      hand *h,*h1;
      
      nhands = ss->hands;
      
      for (i=0; i<nhands-1; i+=2) 
      {
         int  width = (m_across - scoop)/3;
         int  tmp   = width - (width * i) / nhands;
         
         h  = &hands[i];
         h1 = &hands[i+1];
         
         /* Catch on left of hand 0 and on right of hand 1 */
         
         h->throw.x = m_across/2 - tmp;
         h->catch.x = h->throw.x - scoop; /* catch on outside, scoop in */
         
         h1->throw.x = m_across/2 + tmp;
         h1->catch.x = h1->throw.x + scoop;
      }
      
      if (nhands % 2) 
      {
         hands[nhands-1].catch.x = m_across/2;
         hands[nhands-1].throw.x = m_across/2 + scoop;
      }
      
      for(i=0 ;i<nhands; i++)
      {
         h  = &hands[i];
         
         h->catch.y = 200;
         h->throw.y = 200;
         h->catch.z = 0; /*+100*i;*/
         h->throw.z = 0; /*-100*i;*/
         
         h->ball = NULL;
         h->lateral = 0;
         h->shoulder.x = (h->throw.x + h->catch.x)/2;
         h->shoulder.y = h->throw.y + 4*UPPER_ARM/4;
         h->shoulder.z = h->throw.z + 3*LOWER_ARM/3; 
      
         h->dwell = 2;   /* Default value */
      
         h->init_catch = h->catch;
         h->init_throw = h->throw;
      }
   }
   /*}}}  */
   
   /* Guess an appropriate detail level */
   if (ss->detail == 0)
      ss->detail = max(2, 10 - ss->balls);
   
   for (i = 0; i<ss->balls ; i=i+1 ) 
   {
      balls[i].handto     = i%nhands;                  /* ball 0:left, 1:right, 2:left, etc. */
      balls[i].handfrom   = i%nhands;                  /* ball 0:left, 1:right, 2:left, etc. */
      balls[i].left_right = sign(balls[i].handfrom%2); /* queue on outside */
      balls[i].lateral    = -2*i;                      /* ball i will be thrown 2*i throws from now */
      balls[i].throw      = -1;                        /* ball not yet thrown */
   }
      
   ticks = -2*ss->startup_len;
   
   if (cur_ss != ss)
   {
      cur_ss = ss;
      fit_view_to_pattern();
   }
    
   if (prev_ss != NULL)
      redo_screen();
}
/*}}}  */
    
void
init_body()
/*{{{  */
{
   point   pos;
   int     i;
   
   for(i=0; i<NUM_BODY_PARTS; i++)
   {
      init_ap(&body_part[i]);
      identity_tx(body_part[i].tx);
   }
   
   pos.x = hands[0].shoulder.x + 9*BODY_WIDTH/16;
   pos.y = hands[0].shoulder.y - 3*BODY_HEIGHT/8;
   pos.z = hands[0].shoulder.z;
   offset_tx(body_part[BODY_I].tx, &pos);
   
   pos.y += BODY_HEIGHT/2 + HEAD_SIZE/2;
   offset_tx(body_part[HEAD_I].tx, &pos);

   pos.x -= BODY_WIDTH/4;
   pos.y -= BODY_HEIGHT + HEAD_SIZE/2;
   for(i=LEFT_LEG_I; i<LEFT_LEG_I+2; i++)
   {
      rotate_tx(body_part[i].tx, 'x', -PI_BY_2);
      pos.x += (i-LEFT_LEG_I)*BODY_WIDTH/2;
      offset_tx(body_part[i].tx, &pos);
   }
}
/*}}}  */
    
void
set_timer(int val)
/*{{{  */
{
   static struct itimerval it = {0,0,0,0};

   /*
   if ( it.it_value.tv_usec != val )
      printf("Detail = %d, Waiting for %d micro-secs\n",detail,val);
   */
   
   it.it_value.tv_usec = val;

   setitimer(ITIMER_REAL, &it, NULL);
   waiting = TRUE;
}
/*}}}  */

void
reset_timer()
/*{{{  */
{
   waiting = FALSE;
   signal(SIGALRM, reset_timer);
}
/*}}}  */

void
wait_timer()
/*{{{  COMMENT BSD Version*/
/*
{
   /* Enter critical section wrt signal handler
    * on shared variable 'waiting'
    *

   int smask = sigblock(sigmask(SIGALRM));
   
   if (waiting)
   {
      unset_warp_light();
      sigpause(smask);
   }
   else
      set_warp_light();
      
   /* Exit critical section
    *
   sigsetmask(smask);
}
*/
/*}}}  */
/*{{{  Sys V version*/
{
   /* Enter critical section wrt signal handler
    * on shared variable 'waiting'
    */
   sigset_t   newset,oldset;
   
   (void)sigemptyset(&newset);
   (void)sigaddset(&newset, SIGALRM);
   
   (void)sigprocmask(SIG_BLOCK, &newset, &oldset);
   
   if (waiting)
   {
      unset_warp_light();
      sigsuspend(&oldset);
   }
   else
      set_warp_light();
      
   /* Exit critical section
    */
   (void)sigprocmask(SIG_SETMASK, &oldset, NULL);
}
/*}}}  */

int
main(int argc, char *argv[])
/*{{{  */
{
    siteswap         init_ss;
    int              i, j;
    boolean          adjust;
    
    struct timeval   start,end;
    struct timezone  tz;
    int              frametime, overhead;
    
    signal(SIGALRM, reset_timer);
    
    init_graphics(&argc, argv);
    
    m_across = getmaxx();
    m_down   = getmaxy(); 
    
    parse_args(argc, argv, &init_ss);
    load_pattern(&init_ss);

    /*{{{  Init pixmaps*/
    for(i=0; i<MAX_PATTERN_LENGTH; i++)
       init_ap(&balls[i].ap);
    
    for(i=0; i<MAX_NUM_HANDS; i++)
    {
       init_ap(&hands[i].up_ap);
       init_ap(&hands[i].low_ap);
    }
    
    init_body();
    /*}}}  */
    
    jtime = STD_JTIME;
    
    /*
    gettimeofday(&start,&tz);
    gettimeofday(&end,&tz);
    overhead = (end.tv_sec - start.tv_sec)*1000000 + 
                end.tv_usec - start.tv_usec;
    */
    
    setup_uif();   

    while (!done) 
    {
       /*{{{  Possibly switch patterns*/
       if (new_ss != cur_ss)
       {
          adjust = TRUE;
          
          if (new_ss == NULL)
             new_ss = cur_ss; /* i.e. restart */
          
          load_pattern(new_ss);
       }
       /*}}}  */
       
       now = ticks / 2;
       if (now >= cur_ss->pattern_len)
          now = now - cur_ss->pattern_len;
       
       throwsnow = 0; emptyhand = 0; /* used for diagnostics */
    	
       /* calculate state of each ball after current throw */
       for (i = 0; i < cur_ss->balls; i++)
          calc_ball_state(i, now + MAX_PATTERN_LENGTH);
    
       for(i=0; i<nhands; i++)
          calc_hand_state(i, now + MAX_PATTERN_LENGTH);
       
       do_diagnostics();
       
       detail = cur_ss->detail; 
       height = cur_ss->height;
       
       frametime=0;
       
       /* Animation loop - prints intermediate positions */
       for (j = 0; !done && j < detail; j++) 
       /*{{{  */
       {
          set_timer( 4 * jtime * height / (6 *detail) );
          
          /*gettimeofday(&start,&tz);*/
          
           /* check for input events; loop on a pause */
          do 
          {
             process_events();   
          } while (pausing);
       
          /*{{{  Calculate intermediate ball positions*/
          for (i = 0; i < cur_ss->balls; i++) 
          {
             step_ap(&balls[i].ap);
             
             calc_ball_position(i, j);
          }
          /*}}}  */
          /*{{{  and arm positions*/
          if (show_arms)
             for (i=0; i<nhands; i++)
             {
                step_ap(&hands[i].up_ap);
                step_ap(&hands[i].low_ap);
                
                calc_wrist_position(&hands[i], j);
                calc_elbow_position(&hands[i]);
             }
          /*}}}  */
          
          wait_timer();
          /*{{{  COMMENT work out frame rate*/
          /*
          gettimeofday(&end, &tz);
          
          frametime = (end.tv_sec  - start.tv_sec)*1000000 + 
                       end.tv_usec - start.tv_usec - overhead;
             
          printf("Time taken = %d Time requested = %d.\n",
                    frametime,
                    jtime*height/detail);
          */
          /*}}}  */
          
          /*{{{  Display balls on screen*/
          for (i = 0; i < cur_ss->balls; i++) 
          {
             display_ap(&balls[i].ap, onacid);
          }  
          /*}}}  */
          /*{{{  and the arms*/
          if (show_arms)
             for (i=0; i<nhands; i++)
             {
                display_ap(&hands[i].up_ap, onacid);
                display_ap(&hands[i].low_ap, onacid);
             }
          /*}}}  */
          
          flush_graphics();
       
       }
       /*}}}  */
       
       /*{{{  COMMENT frame rate stuff*/
       if (adjust)
       {
          /*
          printf("Total time = %d (expected %d). Frame time = %d (= %d/s)\n",
                 frametime,
                 jtime*height,
                 frametime/detail,
                 100000*detail/frametime);
             
          set_sys_detail( STD_JTIME * detail / frametime);
          adjust = FALSE;
             */
       }
       /*}}}  */
       
       ticks = (ticks + 1);
       if (ticks > 0)
          ticks = ticks % (cur_ss->pattern_len * 4);
    }
    
    closegraph();

    exit(0);
}
/*}}}  */
 
